#!/usr/bin/env python
#-*- coding:utf-8 -*-
#===============================================================================
#  Copyright (C) 2010 Schaefer & Tobies SuC GmbH.
#  Author: Markus Grimm <mgr@schaefer-tobies.de>
#
#  This file is part of Pyontapi.
#
#  Pyontapi is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyontapi is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Pyontapi.  If not, see <http://www.gnu.org/licenses/>.
#
#===============================================================================

""" Abstract base class for all API classes. """

from schtob.pyontapi.generator import GENERIC_TYPEDEFS
from schtob.pyontapi.errors import PyontapiError

GENERIC_TYPES = GENERIC_TYPEDEFS.values()

class NaArgument(object):
    
    """ Argument class for API calls. """
    
    def __init__(self, value, settings):
        self.value = value
        self.name = settings['name']
        self.type = settings['type']
        self.encrypted = settings.get('encrypted', False)
        self.nonempty = settings.get('nonempty', False)
        self.is_optional = settings.get('is-optional', False)
        self.is_array = settings.get('is-array', False)
    
    def is_set(self):
        """ check if value is set """
        if self.type in GENERIC_TYPES:
            if self.type == bool or self.type == int:
                return self.value is not None
            elif self.type == str:
                return self.value is not None and self.value != ''
            else:
                raise PyontapiError('unknown type %s' % self.type)
        elif self.is_array:
            if self.value:
                return True
            return False
        else:
            return self.type.is_set(self.value)
            
    
    def append_to_doc(self, doc, api):
        """ Append argument values to XML query. """
        if not self.is_optional or self.is_set():
            
            element = doc.createElement(self.name)
            if self.type in GENERIC_TYPES:
                
                text = doc.createTextNode(self.get_content())
                element.appendChild(text)
                api.appendChild(element)
            elif self.is_array:
                for value in self.value:
                    child = self.type.append_to_element(doc, value)
                    if child:
                        element.appendChild(child)
                api.appendChild(element)
            else:
                child = self.type.append_to_element(doc, self.value)
                if child:
                    element.appendChild(child)
                    api.appendChild(element)
    
    def get_content(self):
        """ Getter for ``value``. """
        if self.encrypted:
            raise NotImplementedError('encryption is not supported yet')
        return str(self.value)


class NaField(object):
    
    """ Output field class. """
    
    def __init__(self, settings):
        self.name = settings['name']
        self.type = settings['type']
        self.encrypted = settings.get('encrypted', False)
        self.is_array = settings.get('is-array', False)
    
    def get_value(self, parent):
        """ Get value out of response. """
        if self.is_array:
            values = []
            for child in parent.getElementsByTagName(self.type.name):
                value = self.type.get_value(child)
                if value:
                    values.append(value)
            return values
        
        elif self.type in GENERIC_TYPES:
            if not parent.getElementsByTagName(self.name)[0].childNodes:
                return self.get_content('')
            child = parent.getElementsByTagName(self.name)[0].childNodes[0]
            return self.get_content(child.data)
        else:
            child = parent.getElementsByTagName(self.name)[0]
            return self.type.get_value(child)
    
    def get_content(self, value):
        """ Getter for ``value``. """
        if self.type == bool:
            return value == u'true'
        if self.encrypted:
            raise NotImplementedError('encryption is not supported yet')
        return self.type(value)


class NaNamedType(object):
    
    """ Class for complex ONTAPI types. """
    
    def __init__(self, name, elements):
        self.name = name
        self.elements = elements
    
    def append_to_element(self, doc, value):
        """
        Append element to ``doc``.
        
        If this named type is used as an argument, put the values into the 
        XML request.
        """
        entry = doc.createElement(self.name)
        
        if len(self.elements) == 1 and not self.elements[0].name:
            # if only one element is specified and the name of this 
            # element is empty, we need to create a single text node holding
            # the value
            if not value:
                return None
            entry.appendChild(doc.createTextNode(value))
        else:
            for element in self.elements:
                # we assume that ``value`` is a dict and create child elements 
                # for each element in the list
                if element.name in value:
                    child = element.append_to_element(doc, value[element.name])
                    if child:
                        entry.appendChild(child)
        return entry
    
    def get_value(self, parent):
        """
        Get the value for this field out of ``parent``.
        
        If this named type is used within an output field, read the values out
        of the response and return a dictionary.
        """
        if len(self.elements) == 1 and not self.elements[0].name:
            # if only one element is specified and the name of this 
            # element is empty, we need to create a single text node holding
            # the value
            return parent.childNodes[0].data
        
        value = {}
        for element in self.elements:
            value[element.name] = element.get_value(parent)
        return value
    
    def is_set(self, value):
        """ check if value is set """
        if len(self.elements) == 1 and not self.elements[0].name:
            if value == '' or value is None:
                return False
            return True
        else:
            for element in self.elements:
                if element.name in value:
                    return True
            return False
        

class NaTypeElement(object):
    
    """ Class for complex type elements. """
    
    def __init__(self, settings):
        self.name = settings['name']
        self.type = settings['type']
        self.encrypted = settings.get('encrypted', False)
        self.is_array = settings.get('is-array', False)
        self.nonempty = settings.get('nonempty', False)
        self.is_optional = settings.get('is-optional', False)
    
    def append_to_element(self, doc, value):
        """
        Append element to ``doc``.
        
        If this named type element is used as an argument, put the values into 
        the XML request.
        """
        if not self.name:
            return doc.createTextNode(self.get_content(value))
        
        element = doc.createElement(self.name)
        if self.type in GENERIC_TYPES:
            text = doc.createTextNode(self.get_content(value))
            element.appendChild(text)
        elif self.is_array:
            for val in value:
                child = self.type.append_to_element(doc, val)
                element.appendChild(child)
        else:
            child = self.type.append_to_element(doc, value[self.name])
            element.appendChild(child)
        return element
    
    def get_value(self, parent):
        """
        Get the value for this field out of ``parent``.
        
        if this named type element is used within an output field, read the 
        values out of the response.
        """
        children = parent.getElementsByTagName(self.name)
        default = None
        if self.is_array:
            default = []
        
        if not children:
            return default
        
        root = children[0]
        
        if self.is_array:
            values = []
            for child in root.childNodes:
                values.append(self.type.get_value(child))
            return values
            
        if self.type in GENERIC_TYPES:
            if not root.childNodes:
                return default
            value = root.childNodes[0].data
            if self.type == bool:
                return value == u'true'
            return self.type(value)
        
        return self.type.get_value(root)
    
    def get_content(self, value):
        """ Getter for ``value``. """
        if self.encrypted:
            raise NotImplementedError('encryption not supported yet')
        return str(value)


class BaseAPI(object):
    
    """ Base class for all ONTAPI classes. """
    
    __api_name__ = 'base'
    
    def __init__(self, filer):
        self._filer = filer
    
    def add_to_class(cls, filer):
        """ Adds ``cls`` to ``filer`` instance. """
        inst = cls(filer)
        setattr(filer, cls.__api_name__, inst)
        filer.api_modules[cls.__api_name__] = inst

    add_to_class = classmethod(add_to_class)
